//
//	Ten-Tec 1254 CPU
//
//	Version 1.5(̻С)
//
//	MCU:		ATmega8
//	FUSE:		D1E3h(internal CR osc. 4MHz)
//	Compiler:	AVR-GCC
//
//	2003.10.14	KAWAKAMI Yukio
//


#include	<avr/io.h>
#include	<avr/signal.h>
#include	<avr/interrupt.h>
#include	<avr/pgmspace.h>

#define		LF	10
#define		ON	1
#define		OFF	0
#define		TRUE	1
#define		FALSE	0

#define		NOP		asm volatile("nop\n"::)
#define		WDR		asm volatile("wdr\n"::)
#define		SLEEP	asm volatile("sleep\n"::)
#define		RESET	asm volatile("rjmp\t0\n"::)

typedef	unsigned char	uchar;
typedef	unsigned int	uint;
typedef	unsigned long	ulong;


// ݡȳ ͥ褷ΤǥХХ

#define		SEG_A		5	// PD5
#define		SEG_B		7	// PB7
#define		SEG_C		6	// PB6
#define		SEG_D		1	// PD1

#define		RE_B		4	// PB4
#define		RE_A		5	// PB5
#define		ADIN		4	// PC4

#define		POWER_OFF	2	// PD2
#define		ENABLE		3	// PD3
#define		CLK			4	// PD4
#define		DIN			5	// PC5
#define		VCO			3	// PC3
#define		AM			2	// PC2
#define		SSB			1	// PC1
#define		FM			0	// PC0

#define		LED1		3	// PB3
#define		LED2		2	// PB2
#define		LED3		1	// PB1
#define		LED4		0	// PD0
#define		LED5		6	// PD6
#define		LED6		7	// PD7
#define		LED7		0	// PB0

#define		MODE_AM		0
#define		MODE_SSB	1
#define		MODE_MW		2
#define		MODE_FM		3

#define		LED_AM		12	// AM LED ѥ
#define		LED_SSB		10	// SSB LED ѥ
#define		LED_FM		11	//  LEDѥ

#define		FREQ_MIN	1000	// AM/SSB⡼ɤβ
#define		FREQ_MAX	300000	//             
#define		MW_MIN		5220	// Ȥβ(0.1kHzñ)
#define		MW_MAX		16290	//     (        )
#define		FM_MIN		760000	// FM⡼ɤβ

#define		EEPROM_MODE		1		// POWER-OFFΥ⡼
#define		EEPROM_FREQ1_H	2		//            ȿ(AM/SSB)
#define		EEPROM_FREQ1_M	3
#define		EEPROM_FREQ1_L	4
#define		EEPROM_FREQ2_H	5		//            ȿ(MW/FM)
#define		EEPROM_FREQ2_M	6
#define		EEPROM_FREQ2_L	7
#define		EEPROM_MEMORY	8		// ꡼ch

#define		EEPROM_CLOCK	9		// ưå
#define		EEPROM_CH_WAIT	10		// chɽ(100mSñ)
#define		EEPROM_FM_MAX	11		// FM⡼ɻκȿ(MHz)
#define		EEPROM_FM_STEP	12		//           ƥå(kHz)
#define		EEPROM_ENCORDER	13		// ꡼󥳡Υ
#define		EEPROM_FM_IF	14		// FM IFȿ(100kHzñ)
#define		EEPROM_VCO		15		// VCOؤȿ

#define		EEPROM_MEM_START	20	// ꡼ 1ch 4byte  MEM_MAX ch
#define		MEM_MAX				50	// ꡼ο

#define		CONFIG_ENCORDER	1		// ꡼󥳡
#define		CONFIG_CH_WAIT	2		// chɽ
#define		CONFIG_FM_MAX	3		// FM⡼ɻκȿ
#define		CONFIG_FM_STEP	4		//           ƥå
#define		CONFIG_FM_IF	5		//           IFȿ
#define		CONFIG_VCO		6		// VCOؤȿ
#define		CONFIG_CLOCK	7		// ưåĴ
#define		CONFIG_CLEAR	8		// EEPROM륯ꥢ

void	write_eeprom(uint adrs, uchar data);
uchar	read_eeprom(uint adrs);

void	set145170(uchar bit, uint data);
void	set_freq(ulong freq);
void	set_mode(uchar mode);
void	print_freq(ulong freq);
void	print_ch(char ch);
ulong	formal_freq(ulong freq, ulong step);
ulong	read_memory_freq(char ch);
uchar	read_memory_mode(char ch);
void	write_memory(char ch, uchar mode, ulong freq);
void	set_switch(volatile uchar *flag, volatile uchar *old);
uchar	read_encorder(void);

void	config(void);
void	print_config(uchar id, uchar data);
void	config_encorder(uchar menu, uchar com);
void	config_ch_wait(uchar menu, uchar com);
void	config_fm_max(uchar menu, uchar com);
void	config_fm_step(uchar menu, uchar com);
void	config_fm_if(uchar menu, uchar com);
void	config_clock(uchar menu, uchar com);
void	config_clear(uchar menu, uchar com);
void	config_vco(uchar menu, uchar com);

#define		ENC_NOP		0
#define		ENC_UP		1
#define		ENC_DOWN	2
#define		ENTER		3
#define		INIT		4

PROGMEM uchar EncData[] = {		// ꡼󥳡ư
	ENC_NOP,	ENC_UP,		ENC_DOWN,	ENC_NOP,	// 0  0, 1, 2, 3
	ENC_DOWN,	ENC_NOP,	ENC_NOP,	ENC_UP,		// 1  0, 1, 2, 3
	ENC_UP,		ENC_NOP,	ENC_NOP,	ENC_DOWN,	// 2  0, 1, 2, 3
	ENC_NOP,	ENC_DOWN,	ENC_UP,		ENC_NOP		// 3  0, 1, 2, 3
};


ulong	Freq_min;
ulong	Freq_max;
uint	Freq_step;

ulong	Freq;		// ߤμȿ  100Hzñ̤Ǥ뤳Ȥ
uchar	Mode;		// ߤΥ⡼
uchar	Mode_bak;

ulong	Freq_SW;	// AM/SSB⡼ɤμȿ
ulong	Freq_MW;	// MW/FM⡼ɤμȿ

uchar	Fast;
char	Memory;
uchar	Mem_write;	// ꡼񤭹ߥ⡼
uint	Mem_timer;

// EEPROMɤߤͤȤ
uint	CH_wait;	// ͥɽ
ulong	FM_max;		// FM⡼ɻκȿ
uchar	FM_step;	//           ȿƥå
uchar	FM_IF;		//           IFȿ
uchar	Enc_type;	// ꡼󥳡Υ
ulong	change_VCO;	// VCOؤȿ

volatile uchar	SW_mode;
volatile uchar	SW_mode_old;
volatile uchar	SW_mw;
volatile uchar	SW_mw_old;
volatile uchar	SW_vm;
volatile uchar	SW_vm_old;
volatile uchar	SW_speed;
volatile uchar	SW_speed_old;

volatile uchar	LED_scan;
volatile uchar	LED_buff[7];	// LEDɽХåե

volatile uchar	Mode_Bit_buff;	// ߽ MC145170
volatile uint	Mode_Data_buff;	// ǡ򥻥åȤ٤ΥХåե
volatile uchar	Freq_Bit_buff;
volatile uint	Freq_Data_buff;

volatile uchar	Debug;


//*********************************
//	ޡСե
//	4MHz/64/125=500 500Hz(¢CRȯʤΤǸ)
//*********************************
SIGNAL(SIG_OVERFLOW0) {
	uchar	led;
	
	// γ
	outp((uchar)(256-(125-1)), TCNT0);

	// LEDöä
	cbi(PORTB, LED1);
	cbi(PORTB, LED2);
	cbi(PORTB, LED3);
	cbi(PORTD, LED4);
	cbi(PORTD, LED5);
	cbi(PORTD, LED6);
	cbi(PORTB, LED7);

	if (++LED_scan > 6){
		LED_scan = 0;
	}

	if (LED_scan == 0){	// dpɤ٤ˤΥߥ󥰤PLL

		// ⡼ѹм¹
		if (Mode_Bit_buff){
			set145170(Mode_Bit_buff, Mode_Data_buff);
			Mode_Bit_buff = 0;
		}
		// ȿѹм¹
		if (Freq_Bit_buff){
			set145170(Freq_Bit_buff, Freq_Data_buff);
			Freq_Bit_buff = 0;
		}
	}

	// 7SEG
	led = LED_buff[LED_scan];
	if (led & 1){
		sbi(PORTD, SEG_A);
	} else {
		cbi(PORTD, SEG_A);
	}
	if (led & 2){
		sbi(PORTB, SEG_B);
	} else {
		cbi(PORTB, SEG_B);
	}
	if (led & 4){
		sbi(PORTB, SEG_C);
	} else {
		cbi(PORTB, SEG_C);
	}
	if (led & 8){
		sbi(PORTD, SEG_D);
	} else {
		cbi(PORTD, SEG_D);
	}
	if (led & 0x80){
		sbi(PORTD, CLK);
	} else {
		cbi(PORTD, CLK);
	}

	// ϡå
	cbi(DDRC, DIN);
	switch(LED_scan){
	  case 0:
		sbi(PORTB, LED7);
		break;
	  case 1:
		sbi(PORTD, LED6);
		break;
	  case 2:
		sbi(PORTD, LED5);
		break;
	  case 3:
		sbi(PORTD, LED4);
		set_switch(&SW_mode, &SW_mode_old);
		break;
	  case 4:
		sbi(PORTB, LED3);
		set_switch(&SW_mw, &SW_mw_old);
		break;
	  case 5:
		sbi(PORTB, LED2);
		set_switch(&SW_vm, &SW_vm_old);
		break;
	  default:
		sbi(PORTB, LED1);
		set_switch(&SW_speed, &SW_speed_old);
		break;
	}
}


//*********************************
//	å
//*********************************
void set_switch(volatile uchar *flag, volatile uchar *old){
	*old <<= 1;
	if (inp(PINC)&(1<<DIN)){
		*old |= 1;
	}
	if ((*old & 0x0F) == 0){
		*flag = OFF;
	} else if ((*old & 0x0F) == 0x0F){
		*flag = ON;
	}
}


//*********************************
//	INT0(POWER-OFF)ߥ롼
//*********************************
SIGNAL(SIG_INTERRUPT0) {

	// EEPROM񤭹Ϥݤ١LEDݡȽOFF
	outp((1<<SEG_B)|(1<<SEG_C), PORTB);
	outp((1<<SEG_A)|(1<<SEG_D)|(1<<ENABLE), PORTD);
	outp(0, PORTC);

	// ⡼ɤȼȿ EEPROM˽񤭹
	if ((Memory <= 0) || Mem_write){
		if ((Mode == MODE_MW||(Mode == MODE_FM))){
			Freq_MW = Freq;
		} else {
			Freq_SW = Freq;
		}
	}
	if (Memory <= 0){
		write_eeprom(EEPROM_MODE, Mode);
	} else {
		write_eeprom(EEPROM_MODE, Mode_bak);
	}
	write_eeprom(EEPROM_FREQ1_H, Freq_SW >> 16);
	write_eeprom(EEPROM_FREQ1_M, Freq_SW >> 8);
	write_eeprom(EEPROM_FREQ1_L, Freq_SW);
	write_eeprom(EEPROM_FREQ2_H, Freq_MW >> 16);
	write_eeprom(EEPROM_FREQ2_M, Freq_MW >> 8);
	write_eeprom(EEPROM_FREQ2_L, Freq_MW);
	write_eeprom(EEPROM_MEMORY, (uchar)Memory);

	read_eeprom(0);	// ߽λޤԤ

	SLEEP;	// 
}


//*********************************
//	ᥤ롼
//*********************************
int main (void) {
	uint	i;
	uchar	osc, *p;

	cli();		// ߶ػ

	// ATmega8 SRAMꥢ
	p = (uchar *)0x60;
	for (i = 0; i < 1024; i++){
		*p++ = 0;
	}
	for (i = 0; i < 7; i++){
		LED_buff[i] = 0x0F;
	}

	// ݡȤ
	outp((1<<SEG_B)|(1<<SEG_C), PORTB);
	outp((1<<SEG_A)|(1<<SEG_D)|(1<<ENABLE), PORTD);
	outp((1<<ADIN), PORTC);

	outp((1<<SEG_B)|(1<<SEG_C)|(1<<LED1)|(1<<LED2)|(1<<LED3)|(1<<LED7), DDRB);
	outp((1<<SEG_A)|(1<<SEG_D)|(1<<LED6)|(1<<LED5)|(1<<LED4)|(1<<CLK)|(1<<ENABLE), DDRD);
	outp((1<<FM)|(1<<SSB)|(1<<AM)|(1<<VCO), DDRC);
	
	// ŰꤹޤԤ
	for (i = 0; i < 60000;){
		if (inp(PIND)&(1<<POWER_OFF)){
			i = 0;
		} else {
			i++;
		}
	}

	// MC145170ν
	set145170(8, 0xA0);		// POL=H, PDB, LDE=1, REFOUT=off

	// ưåĴ
	osc = read_eeprom(EEPROM_CLOCK);
	if ((osc >= 50)&&(osc <= 200)){
		outp(osc, OSCCAL);
	}

	// EEPROMƼɤ߹
	CH_wait = (uint)read_eeprom(EEPROM_CH_WAIT) * (100/2);
	if (CH_wait > (5000/2))		CH_wait = 1500/2;
	FM_max = (ulong)read_eeprom(EEPROM_FM_MAX) * 10000;
	if (FM_max > 1080000)		FM_max = 900000;
	FM_step = read_eeprom(EEPROM_FM_STEP);
	if ((FM_step != 50)&&(FM_step != 100))	FM_step = 50;
	Enc_type = read_eeprom(EEPROM_ENCORDER);
	if (Enc_type > 1)			Enc_type = 0;
	FM_IF = read_eeprom(EEPROM_FM_IF);
	if ((FM_IF < 100)||(FM_IF == 255))	FM_IF = 107;
	i = read_eeprom(EEPROM_VCO);
	switch(i){
	  case 12:
		change_VCO = 125000;		break;
	  case 14:
		change_VCO = 145000;		break;
	  case 15:
		change_VCO = 159000;		break;
	  case 16:
		change_VCO = 165000;		break;
	  case 17:
		change_VCO = 179500;		break;
	  case 18:
		change_VCO = 185000;		break;
	  case 19:
		change_VCO = 195000;		break;
	  case 20:
		change_VCO = 205000;		break;
	  default:
		change_VCO = 132400;		break;
	}

	// EEPROMȿɤ߹
	Mode = read_eeprom(EEPROM_MODE);
	if (Mode == 0xFF){	// ץ񤭴ľ
		Freq_SW = FREQ_MIN;
		Freq_MW = MW_MIN;
		Freq = Freq_SW;
		Mode = MODE_AM;
		Memory = -1;
	} else {
		Freq_SW = ((ulong)read_eeprom(EEPROM_FREQ1_H) << 16)
			+ ((ulong)read_eeprom(EEPROM_FREQ1_M) << 8)
			+ ((ulong)read_eeprom(EEPROM_FREQ1_L));
		Freq_MW = ((ulong)read_eeprom(EEPROM_FREQ2_H) << 16)
			+ ((ulong)read_eeprom(EEPROM_FREQ2_M) << 8)
			+ ((ulong)read_eeprom(EEPROM_FREQ2_L));
		Memory = (char)read_eeprom(EEPROM_MEMORY);
		if (Memory > 0){
			Mode_bak = Mode;
			Mode = read_memory_mode(Memory);
			Freq = read_memory_freq(Memory);
		} else {
			if ((Mode == MODE_MW)||(Mode == MODE_FM)){
				Freq = Freq_MW;
			} else {
				Freq = Freq_SW;
			}
		}
	}

	outp(0x18, WDTCR);	// WDCEWDE1ˤƥޡѹ򤹤
	outp(0x1F, WDTCR);	// åɥåޡå /5V

	outp(0x83, MCUCR);	// INT0(POWER-OFF)Ω夬,꡼׵
	sbi(GICR, INT0);	// 

	outp((uchar)(256-(125-1)), TCNT0);	// ޡ -125
	outp(0x03, TCCR0);				// ޡץꥹ 1/64
	sbi(TIMSK, TOIE0);				// ޳ߵ

	sei();		// ߵ

	// ưΥåå
	for (i = 0; i < 50; i++){
		SLEEP;
	}
	if (SW_mw){
		config();	// MWܥ󤬲Ƥ⡼
		// ܥޤԤ
		while (SW_mode || SW_mw || SW_vm || SW_speed)	SLEEP;
		RESET;
	}

	set_mode(Mode);
	set_freq(Freq);
	if (Memory > 0){
		print_ch(Memory);	// chɽ
	} else {
		print_freq(Freq);	// ȿɽ
	}

	read_encorder();		// 󥳡

	for (;;){
		// ܥޤԤ
		while (SW_mode || SW_mw || SW_vm || SW_speed)	SLEEP;

		WDR;
		SLEEP;

		// 󥳡ɤ߹
		switch (read_encorder()){

		  case ENC_UP:
			if (Memory > 0){
				if (++Memory > MEM_MAX)	Memory = 1;
				if (!Mem_write){
					Mode = read_memory_mode(Memory);
					Freq = read_memory_freq(Memory);	// PLL
					set_mode(Mode);
					set_freq(Freq);
				}
				print_ch(Memory);	// chɽ
			} else {
				if (Fast){
					Freq += 1000;
					Freq = formal_freq(Freq, 1000);
				} else {
					Freq += (ulong)Freq_step;
				}
				if ((Mode == MODE_MW)&&(Freq > MW_MAX)&&(FM_max != 0)){
					Mode = MODE_FM;
					Freq = FM_MIN;
					set_mode(Mode);
				}
				Freq = formal_freq(Freq, (ulong)Freq_step);
				set_freq(Freq);		// PLL
				print_freq(Freq);	// LEDɽ
			}
			break;

		  case ENC_DOWN:
			if (Memory > 0){
				if (--Memory < 1)	Memory = MEM_MAX;
				if (!Mem_write){
					Mode = read_memory_mode(Memory);
					Freq = read_memory_freq(Memory);	// PLL
					set_mode(Mode);
					set_freq(Freq);
				}
				print_ch(Memory);	// chɽ
			} else {
				if (Fast){
					Freq -= 1000;
					Freq = formal_freq(Freq, 1000);
				} else {
					Freq -= (ulong)Freq_step;
				}
				if ((Mode == MODE_FM)&&(Freq < FM_MIN)){
					Mode = MODE_MW;
					Freq = MW_MAX;
					set_mode(Mode);
				}
				Freq = formal_freq(Freq, (ulong)Freq_step);
				set_freq(Freq);		// PLL
				print_freq(Freq);	// LEDɽ
			}
			break;

		  default:
			// ꡼⡼ɤǥޡڤ줿ȿɽ
			if ((Memory > 0) && Mem_timer && (--Mem_timer == 0)){
				print_freq(Freq);
			}
			break;
		}

		if (SW_mw){		// MWܥ󤬲줿
			if ((Memory <= 0) && !Mem_write){
				Mem_write = TRUE;
				Memory *= -1;
				if (Memory == 0)	Memory = 1;
				if ((Memory < 1)||(Memory > MEM_MAX))	Memory = 1;
				print_ch(Memory);	// chɽ
			} else if (Mem_write){
				Mem_write = FALSE;
				write_memory(Memory, Mode, Freq);	// ꡼EEPROM񤭹
				Memory *= -1;
				print_freq(Freq);	// LEDɽ
			}

		} else if (SW_vm){		// V/Mܥ󤬲Ƥ
			Fast = FALSE;
			if (Mem_write){
				// ꡼񤭹ߥ⡼ɤǲ줿Ͻ񤭹ߥ󥻥
				Mem_write = FALSE;
				Memory *= -1;
				print_freq(Freq);	// LEDɽ
			} else {
				if (Memory <= 0){
					if ((Mode == MODE_MW)||(Mode == MODE_FM)){
						Freq_MW = Freq;
					} else {
						Freq_SW = Freq;
					}
					Mode_bak = Mode;
					Memory *= -1;
					if ((Memory < 1)||(Memory > MEM_MAX))	Memory = 1;
					Mode = read_memory_mode(Memory);
					Freq = read_memory_freq(Memory);	// PLL
					set_mode(Mode);
					set_freq(Freq);
					print_ch(Memory);	// chɽ
				} else {
					Memory *= -1;
					Mode = Mode_bak;
					if ((Mode == MODE_MW)||(Mode == MODE_FM)){
						Freq = Freq_MW;
					} else {
						Freq = Freq_SW;
					}
					set_mode(Mode);
					set_freq(Freq);		// PLL
					print_freq(Freq);	// LEDɽ
				}
			}

		} else if ((Memory <= 0) && SW_speed
				   && ((Mode == MODE_AM)||(Mode == MODE_SSB))){
			// ԡɥܥ󤬲줿
			Fast ^= 1;
			if (Fast){
				LED_buff[6] |= 0x80;
			} else {
				LED_buff[6] &= 0x7F;
			}
			
		} else if ((Memory <= 0) && SW_mode){	// MODEܥ󤬲줿
			Fast = FALSE;
			if (Mode == MODE_AM){
				Mode = MODE_SSB;
				Freq_SW = Freq;
			} else if (Mode == MODE_SSB){
				Freq_SW = Freq;
				Freq = Freq_MW;
				if (Freq > MW_MAX){
					Mode = MODE_FM;
				} else {
					Mode = MODE_MW;
				}
			} else {
				Freq_MW = Freq;
				Freq = Freq_SW;
				Mode = MODE_AM;
			}
			set_mode(Mode);
//			Freq = formal_freq(Freq, Freq_step);
			set_freq(Freq);		// PLL
			print_freq(Freq);	// LEDɽ

		}
	}
}


//*********************************
//	⡼
//*********************************
void config(void){
	uchar	com;
	uchar	menu;

	read_encorder();	// 󥳡

	menu = 1;
	config_encorder(menu, INIT);	// ɽ

	for (;;){

		// ܥޤԤ
		while (SW_mode || SW_mw || SW_vm || SW_speed)	SLEEP;

		WDR;
		SLEEP;

		com = ENC_NOP;
		if (SW_mw){				// [MW] ǡ񤭤
			com = ENTER;
		} else if (SW_mode){	// [MODE] Υ˥塼
			if (++menu > CONFIG_CLEAR) menu = 1;
			com = INIT;
		} else if (SW_vm){		// [V/M] ⡼ɤȴ
			read_eeprom(0);	// EEPROM߽λޤԤ
			return;
		}

		if (com == ENC_NOP){
			// 󥳡ɤ߹
			com = read_encorder();
		}

		if (com != ENC_NOP){
			switch(menu){
			  case CONFIG_ENCORDER:
				config_encorder(menu, com);		break;
			  case CONFIG_CH_WAIT:
				config_ch_wait(menu, com);		break;
			  case CONFIG_FM_MAX:
				config_fm_max(menu, com);		break;
			  case CONFIG_FM_STEP:
				config_fm_step(menu, com);		break;
			  case CONFIG_FM_IF:
				config_fm_if(menu, com);		break;
			  case CONFIG_VCO:
				config_vco(menu, com);			break;
			  case CONFIG_CLOCK:
				config_clock(menu, com);		break;
			  case CONFIG_CLEAR:
				config_clear(menu, com);		break;
			  default:
				menu = 1;						break;
			}
		}
	}
}


//*********************************
//	ưåĴ
//*********************************
void config_clock(uchar menu, uchar com){
	uchar	osc;

	osc = inp(OSCCAL);
	if (com == ENC_UP){
		if (osc < 200)	osc++;
		outp(osc, OSCCAL);
	} else if (com == ENC_DOWN){
		if (osc > 50)	--osc;
		outp(osc, OSCCAL);
	} else if (com == ENTER){
		write_eeprom(EEPROM_CLOCK, osc);
	} else if (com == INIT){
		osc = read_eeprom(EEPROM_CLOCK);
		if ((osc >= 50)&&(osc <= 200)){
			outp(osc, OSCCAL);
		} else {
			osc = inp(OSCCAL);
		}
	}

	print_config(menu, osc);
}


//*********************************
//	chɽ
//*********************************
void config_ch_wait(uchar menu, uchar com){
	static uchar	tm = 15;

	if (com == ENC_UP){
		if (tm < 50)	tm++;
	} else if (com == ENC_DOWN){
		if (tm > 0)		--tm;
	} else if (com == ENTER){
		write_eeprom(EEPROM_CH_WAIT, tm);
	} else if (com == INIT){
		tm = read_eeprom(EEPROM_CH_WAIT);
		if (tm > 50){
			tm = 15;
		}
	}

	print_config(menu, tm);
	if (LED_buff[4] == 0x0F){
		LED_buff[4] = 0;
	}
	LED_buff[4] |= 0x80;	// ɽ
}


//*********************************
//	FM⡼ɾ¼ȿ
//*********************************
void config_fm_max(uchar menu, uchar com){
	static uchar	freq = 0;

	if (com == ENC_UP){
		if (freq < 108)		freq++;
		if (freq < 77)		freq = 77;
	} else if (com == ENC_DOWN){
		if (freq > 0)		--freq;
		if (freq < 77)	freq = 0;
	} else if (com == ENTER){
		write_eeprom(EEPROM_FM_MAX, freq);
	} else if (com == INIT){
		freq = read_eeprom(EEPROM_FM_MAX);
		if (freq < 77)	freq = 0;
		if (freq > 108)	freq = 90;
	}

	print_config(menu, freq);
}


//*********************************
//	FM⡼ɥƥå
//*********************************
void config_fm_step(uchar menu, uchar com){
	static uchar	freq = 50;

	if (com == ENC_UP){
		freq = 100;
	} else if (com == ENC_DOWN){
		freq = 50;
	} else if (com == ENTER){
		write_eeprom(EEPROM_FM_STEP, freq);
	} else if (com == INIT){
		freq = read_eeprom(EEPROM_FM_STEP);
		if ((freq != 50)&&(freq != 100)){
			freq = 50;
		}
	}

	print_config(menu, freq);
}


//*********************************
//	FM⡼ IFȿ
//*********************************
void config_fm_if(uchar menu, uchar com){
	static uchar	freq = 107;

	if (com == ENC_UP){
		if (freq < 254)		freq++;
	} else if (com == ENC_DOWN){
		if (freq > 100)		--freq;
	} else if (com == ENTER){
		write_eeprom(EEPROM_FM_IF, freq);
	} else if (com == INIT){
		freq = read_eeprom(EEPROM_FM_IF);
		if ((freq < 100)||(freq == 255)){
			freq = 107;
		}
	}

	print_config(menu, freq);
	LED_buff[4] |= 0x80;
}


//*********************************
//	꡼󥳡
//*********************************
void config_encorder(uchar menu, uchar com){
	static uchar	sel = 0;

	if (com == ENC_UP){
		sel = 1;
	} else if (com == ENC_DOWN){
		sel = 0;
	} else if (com == ENTER){
		write_eeprom(EEPROM_ENCORDER, sel);
		Enc_type = sel;	// ȿ
	} else if (com == INIT){
		sel = read_eeprom(EEPROM_ENCORDER);
		if (sel > 1){
			sel = 0;
		}
	}

	if (sel == 0){
		print_config(menu, 1);	// ǥե
	} else {
		print_config(menu, 4);	// 
	}
}


//*********************************
//	EEPROM 륯ꥢ
//*********************************
void config_clear(uchar menu, uchar com){
	static uchar	clr = 99;

	if (com == ENC_UP){
		if (clr < 99)	clr++;
	} else if (com == ENC_DOWN){
		if (clr > 0)	--clr;
	} else if (com == ENTER){
		if (clr == 0){
			int	i;
			for (i = 0; i < 512; i++){
				write_eeprom(i, 0xFF);
			}
		}
	} else if (com == INIT){
		clr = 99;
	}

	print_config(menu, clr);
}


//*********************************
//	VCOؤȿ
//*********************************
void config_vco(uchar menu, uchar com){
	static uchar	freq = 13;

	if (com == ENC_UP){
		if (freq < 20)		freq++;
	} else if (com == ENC_DOWN){
		if (freq > 12)		--freq;
	} else if (com == ENTER){
		write_eeprom(EEPROM_VCO, freq);
	} else if (com == INIT){
		freq = read_eeprom(EEPROM_VCO);
		if ((freq < 12)||(freq > 20)){
			freq = 13;
		}
	}

	print_config(menu, freq);
}


//*********************************
//	⡼
//*********************************
void set_mode(uchar	m){

	cli();					// ߶ػ
	switch (m){
	  case MODE_SSB:
		cbi(PORTC, FM);
		cbi(PORTC, AM);
		// ȿ 3.58MHz/1432 = 2.5kHz
		Mode_Bit_buff = 15;
		Mode_Data_buff = 1432;
		Freq_min = FREQ_MIN;
		Freq_max = FREQ_MAX;
		Freq_step = 25;
		LED_buff[6] = LED_SSB;
		sbi(PORTC, SSB);
		break;

	  case MODE_MW:
		cbi(PORTC, FM);
		cbi(PORTC, SSB);
		// ȿ 3.58MHz/3580 = 1kHz
		Mode_Bit_buff = 15;
		Mode_Data_buff = 3580;
		Freq_min = MW_MIN;
		Freq_max = MW_MAX;
		Freq_step = 90;
		LED_buff[6] = LED_AM;
		sbi(PORTC, AM);
		break;

	  case MODE_FM:
		cbi(PORTC, AM);
		cbi(PORTC, SSB);
		// ȿ 3.58MHz/1432 = 2.5kHz
		Mode_Bit_buff = 15;
		Mode_Data_buff = 1432;
		Freq_min = FM_MIN;
		Freq_max = FM_max;
		Freq_step = (uint)FM_step * 10;
		LED_buff[6] = LED_FM;
		sbi(PORTC, FM);
		break;

	  default:
		cbi(PORTC, FM);
		cbi(PORTC, SSB);
		// ȿ 3.58MHz/1432 = 2.5kHz
		Mode_Bit_buff = 15;
		Mode_Data_buff = 1432;
		Freq_min = FREQ_MIN;
		Freq_max = FREQ_MAX;
		Freq_step = 50;
		LED_buff[6] = LED_AM;
		sbi(PORTC, AM);
		break;
	}
	sei();					// ߵ
}


//*********************************
//	ȿ
//*********************************
ulong formal_freq(ulong freq, ulong step){
	ulong	mod;

	if (freq < Freq_min){
		return Freq_min;
	} else if (freq > Freq_max){
		return Freq_max;
	}

	mod = freq % step;
	if (mod == 0){
		return freq;
	}

	if (mod >= (step / 2)){
		freq += step - mod;
	} else {
		freq -= mod;
	}
	return freq;
}


//*********************************
//	PLL ȿ
//*********************************
void set_freq(ulong freq){

	if (freq < FM_MIN){
		// ûӰʲ
		if (freq >= change_VCO){
			cbi(PORTC, VCO);
		} else {
			sbi(PORTC, VCO);
		}
		freq += 450000;		// 1st IF = 45MHz
		if (Mode == MODE_MW){
			freq /= 10;		// ȿ 1kHz
		} else {
			freq /= 25;		// ȿ 2.5kHz
		}
	} else {
		// FMХ
		cbi(PORTC, VCO);
		freq -= (ulong)FM_IF * 1000;	// IF
		freq /= 25;						// PLLȿ 2.5kHz
	}

	Freq_Bit_buff = 16;
	Freq_Data_buff = (uint)freq;
}


//*********************************
//	ȿ LEDɽ
//*********************************
void print_freq(ulong freq){
	uchar	i;
	uchar	buff[6];

	if (freq < FM_MIN){
		// ûӰʲ
		for (i = 5; i != 0xFF; --i){
			buff[i] = (uchar)(freq % 10);
			freq /= 10;
		}
		if (((Memory <= 0)&&(Mode == MODE_MW))
			||((Memory > 0)
			   && !Mem_write && (read_memory_mode(Memory) == MODE_MW))
			||((Memory > 0) && Mem_write && (Mode == MODE_MW))){
			if (buff[1] == 0){
				buff[1] = 0x0f;
			}
		} else {
			buff[1] |= 0x80;
		}

		if (buff[0] == 0){
			buff[0] = 0x0f;
		}
		buff[4] |= 0x80;
	} else {
		// FMХ
		freq /= 10;
		for (i = 5; i != 0xFF; --i){
			buff[i] = (uchar)(freq % 10);
			freq /= 10;
		}
		if (buff[0] == 0){
			buff[0] = 0x0f;
		}
		buff[2] |= 0x80;
	}

	if (Memory <= 0){
		buff[5] |= 0x80;	// 쥯ȥ塼˥󥰻Ϻǲdp
	}

	// LEDХåեž
	for (i = 0; i <= 5; i++){
		LED_buff[i] = buff[i];
	}
}


//*********************************
//	⡼ɽ
//*********************************
void print_config(uchar id, uchar data){
	uchar	i;
	uchar	buff[6];

	buff[0] = id / 10;
	if (buff[0] == 0)	buff[0] = 0x0F;
	buff[1] = (id % 10) | 0x80;
	buff[2] = 0x0F;

	buff[3] = data / 100;
	if (buff[3] == 0)	buff[3] = 0x0F;
	data = data % 100;
	buff[4] = data / 10;
	if ((buff[4] == 0)&&(buff[3] == 0x0F))	buff[4] = 0x0F;
	buff[5] = data % 10;

	// LEDХåեž
	for (i = 0; i <= 5; i++){
		LED_buff[i] = buff[i];
	}
}


//*********************************
//	꡼ch LEDɽ
//*********************************
void print_ch(char ch){

	LED_buff[0] = 0x0F;
	LED_buff[1] = 0x0F;
	LED_buff[2] = 0x0F;
	LED_buff[3] = 0x0A;		// c ɽ
	LED_buff[4] = ch / 10;
	LED_buff[5] = ch % 10;
	if (read_memory_freq(ch) != FREQ_MIN){
		LED_buff[5] |= 0x80;	// ꡼ѤߤʾϺǲdp
	}
	Mem_timer = CH_wait;	// ɽޡå
}


//*********************************
//	MC145170ޥɽ
//	߶ػ߾֤Ǽ¹Ԥ뤳
//*********************************
void set145170(uchar bit, uint data){
	uchar	i;

	cbi(PORTD, CLK);		// CLK = L
	sbi(DDRC, DIN);			// DIN = out
	cbi(PORTD, ENABLE);		// ENABLE = L

	for (i = 16; i > 0; --i){
		if (i <= bit){
			if (data >= 0x8000){
				sbi(PORTC, DIN);
			} else {
				cbi(PORTC, DIN);
			}
			sbi(PORTD, CLK);
			cbi(PORTD, CLK);
		}
		data <<= 1;
	}

	sbi(PORTD, ENABLE);		// ENABLE = H
	cbi(PORTC, DIN);
	cbi(DDRC, DIN);			// DIN = in
}


//*********************************
//	꡼ȿɤ߹
//*********************************
ulong read_memory_freq(char ch){
	ulong	f;

	ch = (ch - 1) * 4 + EEPROM_MEM_START + 1;
	f = (ulong)read_eeprom(ch++) << 16;
	f += (ulong)read_eeprom(ch++) << 8;
	f += (ulong)read_eeprom(ch);
	if (f >= 0xFFFFFF){
		return FREQ_MIN;
	} else {
		return f;
	}
}


//*********************************
//	꡼⡼ɤ߹
//*********************************
uchar read_memory_mode(char ch){
	uchar	mode;

	ch = (ch - 1) * 4 + EEPROM_MEM_START;
	mode = read_eeprom(ch);
	if (mode == 0xFF){
		return MODE_AM;
	} else {
		return mode;
	}
}


//*********************************
// ꡼󥳡ɤ߹
//*********************************
uchar read_encorder(void){
	static uchar	enc = 0;

	enc = ((enc & 3) << 2) | ((inp(PINB) & 0x30)>>4);

	if (!Enc_type){
		return PRG_RDB(&EncData[enc]);
	} else {
		if (enc == 0x01){
			return ENC_UP;
		} else if (enc == 0x0B){
			return ENC_DOWN;
		} else {
			return ENC_NOP;
		}
	}
}


//*********************************
//	꡼񤭹
//*********************************
void write_memory(char ch, uchar mode, ulong freq){

	ch = (ch - 1) * 4 + EEPROM_MEM_START;
	write_eeprom(ch++, mode);
	write_eeprom(ch++, freq >> 16);
	write_eeprom(ch++, freq >> 8);
	write_eeprom(ch, freq);
}


//*******************************************************
//	EEPROM 񤭹
//*******************************************************
void write_eeprom(uint adrs, uchar data){

	// ƱƤʤ񤭹ưԤʤ
	if (read_eeprom(adrs) != data){
		while(inp(EECR) & (1<<EEWE));
		outp(adrs >> 8, EEARH);
		outp(adrs, EEARL);
		outp(data, EEDR);
		sbi(EECR, EEMWE);
		sbi(EECR, EEWE);
	}
}


//*******************************************************
//	EEPROM ɤ߹
//*******************************************************
uchar read_eeprom(uint adrs){

	while(inp(EECR) & (1<<EEWE));
	outp(adrs >> 8, EEARH);
	outp(adrs, EEARL);
	sbi(EECR, EERE);
	return inp(EEDR);
}


